--*******************************************************************************************************************************
--
-- Description: This module performs all PCI Express write/read operations to registers not associated with a specific channel.
--				The AXI_WIDTH constant determines whether slave_dataout is 64-bit or 128-bit aligned.
--					
-- Rev History: 09-04-2013:	Inital release.
--
--*******************************************************************************************************************************

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.model_specific_package.all;
use work.sigfpga_package.all;

entity bar2_common is
port (                                                                                                                              
    -- Reset and clock.                                                                                      
	rst    				: in std_logic;							-- Active high asynchronous reset.
	clk		        	: in std_logic;							-- Bus clock.
	
    -- PCI Express write transaction interface.
	slv_addr 			: in std_logic_vector(15 downto 2);				-- Refer to the PLDA EZDMA2 reference manual.
    slv_dataout     	: in std_logic_vector(AXI_WIDTH-1 downto 0);	-- Refer to the PLDA EZDMA2 reference manual.
	bytevalid 		: in std_logic_vector(1 downto 0);					-- Logic one indicates that all PCI byte valids are active.
	bar2_write 			: in std_logic;									-- Logic one indicates that the current PCI write request is directed at BAR2.
                                                                                                                              
    -- PCI Express read transaction interface.
	bar2_slave_read		: in std_logic;									-- Logic one indicates that the current PCI read request is directed at BAR2.
	bar2_rddata_common	: out std_logic_vector(AXI_WIDTH-1 downto 0);	-- Data return from read operation on BAR2 common register.
    dma_rdaddr      	: in std_logic_vector(15 downto 3);				-- Refer to the PLDA EZDMA2 reference manual.

	-- QDR II+ SRAM control and status signals.
	qdr_wr_data			: out DUAL_72;							-- Data bus carrying four 18-bit QDR II+ SRAM write data values.
	qdr_bw_n			: out DUAL_8;							-- Eight byte enables assigned to the QDR II+ SRAM write data values.
	qdr_rd_data			: in DUAL_72;							-- Data bus carrying four 18-bit QDR II+ SRAM read data values.
	qdr_addr			: out DUAL_QDR;							-- Address selected for current operation.
	qdr_cmd				: out DUAL_2;							-- Single period write (0) and read (1) command strobes in the PCI clock domain.
	qdr_rd_valid		: in DUAL_1;							-- Logic one indicates that read data is available from the QDR II+ SRAM.
    
	-- Command, status and error signals.
	soft_trig			: out std_logic;						-- Software trigger requested over the PCIe bus.
	led_control 		: out std_logic_vector(1 downto 0);  	-- Logic one illuminates each of the two LEDs.
	address_error 		: out std_logic_vector(1 downto 0)		-- Logic one indicates an invalid address for this decoder.
);                                                                                                                     	
end bar2_common;                                                                                                      	   
                                                                                                                               	 
architecture structural of bar2_common is
                                                                                                                               	
	signal led_control_reg	: std_logic_vector(1 downto 0);		-- Mask bits to disable specific sources from illuminating the LEDs.
	signal qdr_bw_n_reg		: DUAL_8;							-- Eight active low byte enables assigned to the QDR II+ SRAM write data values.
	signal qdr_addr_reg		: DUAL_QDR;							-- Address selected for current operation.

begin
	
--***************************************************** WRITE TRANSACTIONS *****************************************************

   process (clk,rst)
    begin
        if rst='1' then
			address_error(0) <= '0';
			soft_trig <= '0';
        	led_control_reg <= (others => '0');
			for i in 0 to 1 loop
        		qdr_wr_data(i) <= (others => '0');
        		qdr_bw_n_reg(i) <= (others => '0');
        		qdr_addr_reg(i) <= (others => '0');
				qdr_cmd(i) <= (others => '0');
			end loop;

        elsif rising_edge(clk) then

            -- Target write decoder that loads slave data into the selected register for a target write completion.
            if bar2_write = '1' and slv_addr(15 downto 12) = x"0" then
                case slv_addr(15 downto 3) & "000" is
                
                	-- x"000" is a read only register.
					
					-- x"010" is defined below for single period trigger.
                    when x"0010" =>
                    	null;
						
					-- This is a 2-bit register used to control LED illumination.
                    when x"0020" =>
						if bytevalid(0) = '1' then led_control_reg <= slv_dataout(1 downto 0); end if;
						
					-- Two lower 18-bit QDR II+ SRAM-A write data words.
                    when x"0030" =>
						if bytevalid(1) = '1' then qdr_wr_data(0)(35 downto 18) <= slv_dataout(49 downto 32); end if;
						if bytevalid(0) = '1' then qdr_wr_data(0)(17 downto 0) <= slv_dataout(17 downto 0); end if;
						
					-- Two upper 18-bit QDR II+ SRAM-A write data words.
                    when x"0038" =>
						if bytevalid(1) = '1' then qdr_wr_data(0)(71 downto 54) <= slv_dataout(49 downto 32); end if;
						if bytevalid(0) = '1' then qdr_wr_data(0)(53 downto 36) <= slv_dataout(17 downto 0); end if;
						
					-- Address and byte enables for QDR II+ SRAM-A.
                    when x"0040" =>
						if bytevalid(1) = '1' then qdr_addr_reg(0) <= slv_dataout(QDR_ADDR_WIDTH+31 downto 32); end if;
						if bytevalid(0) = '1' then qdr_bw_n_reg(0) <= slv_dataout(7 downto 0); end if;
						
					-- Read/write command strobes to QDR II+ SRAM-A.
                    when x"0048" =>
						if bytevalid(0) = '1' then
							if slv_dataout(0) = '0' then
								-- Issue a one clock period write command.
								qdr_cmd(0)(0) <= '1';
							else
								-- Issue a one clock period read command.
								qdr_cmd(0)(1) <= '1';
							end if;
						end if;
						
					-- Two lower 18-bit QDR II+ SRAM-B write data words.
                    when x"0050" =>
						if bytevalid(1) = '1' then qdr_wr_data(1)(35 downto 18) <= slv_dataout(49 downto 32); end if;
						if bytevalid(0) = '1' then qdr_wr_data(1)(17 downto 0) <= slv_dataout(17 downto 0); end if;
						
					-- Two upper 18-bit QDR II+ SRAM-B write data words.
                    when x"0058" =>
						if bytevalid(1) = '1' then qdr_wr_data(1)(71 downto 54) <= slv_dataout(49 downto 32); end if;
						if bytevalid(0) = '1' then qdr_wr_data(1)(53 downto 36) <= slv_dataout(17 downto 0); end if;
						
					-- Address and byte enables for QDR II+ SRAM-B.
                    when x"0060" =>
						if bytevalid(1) = '1' then qdr_addr_reg(1) <= slv_dataout(QDR_ADDR_WIDTH+31 downto 32); end if;
						if bytevalid(0) = '1' then qdr_bw_n_reg(1) <= slv_dataout(7 downto 0); end if;
						
					-- Read/write command strobes to QDR II+ SRAM-B.
                    when x"0068" =>
						if bytevalid(0) = '1' then
							if slv_dataout(0) = '0' then
								-- Issue a one clock period write command.
								qdr_cmd(1)(0) <= '1';
							else
								-- Issue a one clock period read command.
								qdr_cmd(1)(1) <= '1';
							end if;
						end if;
						
					-- Return an error if the address is not valid.
                    when others  =>
						address_error(0) <= '1';
						
                end case;    
			else
				address_error(0) <= '0';
				qdr_cmd(0) <= (others => '0');
				qdr_cmd(1) <= (others => '0');
            end if; 
            
            -- Generate a single period software trigger when there is a write to address x"0010".
			if bar2_write = '1' and bytevalid(0) = '1' and (slv_addr(15 downto 3) & "000" = x"0010") then
				soft_trig <= '1';
			else
				soft_trig <= '0';
			end if;
			
        end if; 
    end process;
    
--***************************************************** READ TRANSACTIONS ******************************************************

    process (clk,rst)
    begin
        if rst = '1' then
            bar2_rddata_common <= (others => '0');
			address_error(1) <= '0';
        elsif rising_edge(clk) then
  
            -- Target read decoder that loads selected register contents onto the DMA bus for a target read completion.
            if bar2_slave_read = '1' and slv_addr(15 downto 12) = x"0" then
            
            	-- Set the default value to zero for cases that do not assign all of the bits.
				bar2_rddata_common <= (others => '0');
				
                case dma_rdaddr(15 downto 3) & "000" is
                
					-- Model specific feature identification.
					when x"0000" =>
						bar2_rddata_common(51 downto 48) <= std_logic_vector(to_unsigned(TX_CHANNEL_COUNT,4));
						bar2_rddata_common(47 downto 44) <= std_logic_vector(to_unsigned(RX_CHANNEL_COUNT,4));
						bar2_rddata_common(43 downto 32) <= MODEL;
						bar2_rddata_common(31 downto 0) <= REVISION;
						
                	--  x"010" is a write only register.
						
					-- This is a 2-bit register used to control LED illumination.
                    when x"0020" =>
						bar2_rddata_common(1 downto 0) <= led_control_reg;
                    
					-- Two lower 18-bit QDR II+ SRAM-A read data words.
					when x"0030" =>
						bar2_rddata_common(49 downto 32) <= qdr_rd_data(0)(35 downto 18);
						bar2_rddata_common(17 downto 0) <= qdr_rd_data(0)(17 downto 0);
											
					-- Two upper 18-bit QDR II+ SRAM-A read data words.
					when x"0038" =>
						bar2_rddata_common((AXI_WIDTH-64)+49 downto (AXI_WIDTH-64)+32) <= qdr_rd_data(0)(71 downto 54);
						bar2_rddata_common((AXI_WIDTH-64)+17 downto (AXI_WIDTH-64)) <= qdr_rd_data(0)(53 downto 36);
										
					-- Address and byte enables for QDR II+ SRAM-A.
					when x"0040" =>
						bar2_rddata_common(QDR_ADDR_WIDTH+31 downto 32) <= qdr_addr_reg(0);
						bar2_rddata_common(7 downto 0) <= qdr_bw_n_reg(0);
										
					-- Read command status from QDR II+ SRAM-A.
					when x"0048" =>
						bar2_rddata_common(AXI_WIDTH-64) <= qdr_rd_valid(0);
						bar2_rddata_common <= (others => '0');
										
					-- Two lower 18-bit QDR II+ SRAM-B read data words.
					when x"0050" =>
						bar2_rddata_common(49 downto 32) <= qdr_rd_data(1)(35 downto 18);
						bar2_rddata_common(17 downto 0) <= qdr_rd_data(1)(17 downto 0);
											
					-- Two upper 18-bit QDR II+ SRAM-B read data words.
					when x"0058" =>
						bar2_rddata_common((AXI_WIDTH-64)+49 downto (AXI_WIDTH-64)+32) <= qdr_rd_data(1)(71 downto 54);
						bar2_rddata_common((AXI_WIDTH-64)+17 downto (AXI_WIDTH-64)) <= qdr_rd_data(1)(53 downto 36);
										
					-- Address and byte enables for QDR II+ SRAM-B.
					when x"0060" =>
						bar2_rddata_common(QDR_ADDR_WIDTH+31 downto 32) <= qdr_addr_reg(1);
						bar2_rddata_common(7 downto 0) <= qdr_bw_n_reg(1);
										
					-- Read command status from QDR II+ SRAM-B.
					when x"0068" =>
						bar2_rddata_common(AXI_WIDTH-64) <= qdr_rd_valid(1);
										
					-- Return an error if the address is not valid.
                    when others  =>                                                                                                 
						bar2_rddata_common(31 downto 0) <= x"badaddb2";
						address_error(1) <= '1';

	            end case;
			else
				address_error(1) <= '0';
            end if;
                        
        end if;
    end process;

--****************************************************** PORT ASSIGNMENTS ******************************************************

    led_control <= led_control_reg;
	SRAM_INDEX : for i in 0 to 1 generate
		qdr_addr(i) <= qdr_addr_reg(i);
		qdr_bw_n(i) <= qdr_bw_n_reg(i);
	end generate;
		
end structural;
