--******************************************************************************************************************************
--
-- Description: This controller performs asychronous read/write transactions to the BPI configuration flash.
--					
-- Rev History: 01-29-2014:	Initial release.
--
--******************************************************************************************************************************

library ieee;
use ieee.std_logic_1164.all;
use IEEE.numeric_std.all;
use work.sigfpga_package.all;

entity bpi_flash is
	port (
    -- Reset and clock.                                                                                      
	rst 			: in std_logic;								-- Active high asynchronous reset.
	clk 			: in std_logic;								-- Bus clock.

	-- Connections to command/status registers in the BAR0 module.
	address			: in std_logic_vector(22 downto 0);			-- Address bus.
	revision_select	: in std_logic_vector(1 downto 0); 			-- Revision address bits.
	revision_enable	: in std_logic; 							-- Active high revision select pin enables revision address bits.
	bpi_wr_data		: in std_logic_vector(15 downto 0);			-- Data input (write) bus.
	bpi_rd_data		: out std_logic_vector(15 downto 0);		-- Data output (read) bus.
	bpi_rd_wrn		: in std_logic; 							-- Read (1) or write (0) transaction select.
	program_voltage	: in std_logic; 							-- Active high enable voltage for program and erase operations.
	bpi_start		: in std_logic;								-- Logic one initiates a BPI bus transaction.
	bpi_busy		: out std_logic; 							-- Logic one indicates that the process is active.
	
	--	The following signals connect to external devices through FPGA pins.
	bpi_a			: out std_logic_vector(BPI_ADD-1 downto 0);	-- Address bus.
	bpi_rs			: out std_logic_vector(1 downto 0);			-- Revision address.
	bpi_rs_disable	: out std_logic; 							-- Revision select disable.
	bpi_d_in		: in std_logic_vector(15 downto 0);			-- Data input (read) bus.
	bpi_d_out		: out std_logic_vector(15 downto 0);		-- Data output (write) bus.
	bpi_d_tristate	: out std_logic;							-- Active high data bus tristate.
	bpi_adv_b		: out std_logic; 							-- Active low address valid.
	bpi_fcs_b		: out std_logic; 							-- Active low chip enable.
	bpi_fwe_b		: out std_logic; 							-- Active low write enable.
	bpi_foe_b		: out std_logic; 							-- Active low output enable.
	bpi_vpp			: out std_logic 							-- Active high enable voltage for program and erase operations.
	);
end bpi_flash;

architecture structural of bpi_flash is

	type flashstate is (idle, add, wr, bus_turn, rd);			-- Type definitions for the flash state machine.
	signal bpi_state	: flashstate;  				-- Current serial machine state.
	signal last_state	: flashstate;  				-- Last serial machine state.
	signal timer		: unsigned(4 downto 0);		-- Counter used to time transaction events.
	
begin
  	
--***************************************************** TRANSACTION TIMER ******************************************************

	-- Create a timer to meet the minimum pulse width on active control signals.
	process (clk,rst)
	begin
		if rst = '1' then
			timer <= (others => '0'); 
		   	
		elsif rising_edge(clk) then
		
			-- Reset the timer during the idle state.
			if bpi_state = idle then
				timer <= (others => '0'); 
			-- Increment the timer during any active state.
			else
				timer <= timer + "00001";
			end if;
			
		end if;
	end process;
	
--************************************************* STATE MACHINE TRANSITIONS **************************************************

	-- Create a finite state machine.
	process (clk,rst)
	begin
		if rst = '1' then
			bpi_state <= idle;
		   	
		elsif rising_edge(clk) then
		
			-- Define each of the state machine transitions.
			case bpi_state is
			
				-- Remain idle while waiting for the application logic to request a flash transaction.
				when idle =>
					-- This state will meet the minimum pulse width required to disable the control lines (8 ns).
					if bpi_start = '1' then
						bpi_state <= add;
					end if;
		
				-- Latch the requested address, this state also meets the requirment to toggle the address valid between rd/wr transactions.
				when add =>
					-- Wait >= 7 ns (@ 250 MHz clock).
					if timer = "00010" then
						if bpi_rd_wrn = '0' then
							bpi_state <= wr;
						else
							bpi_state <= bus_turn;
						end if;
					end if;
		
				-- Process a write transaction.
				when wr =>
					-- Wait >= 40 ns (@ 250 MHz clock).
					if timer = "10000" then
						bpi_state <= idle;
					end if;
		
				-- Tristate both data bus drivers for a single clock turn around cycle.
				when bus_turn =>
					if last_state = add then
						bpi_state <= rd;
					else
						bpi_state <= idle;
					end if;
		
				-- Process a read transaction.
				when rd =>
					-- Wait >= 96 ns (@ 250 MHz clock).
					if timer = "11100" then
						bpi_state <= bus_turn;
					end if;
		
				-- Define the default state to be idle.
				when others =>
					bpi_state <= idle;
				
			end case;
		end if;
	end process; 

--*************************************************** STATE MACHINE EVENTS *****************************************************

	-- Define the events that cause a state machine transition.
	process (clk,rst)
	begin
		if rst = '1' then
				bpi_adv_b <= '1';
				bpi_fwe_b <= '1';
				bpi_foe_b <= '1';
				bpi_fcs_b <= '1';
				bpi_d_tristate <= '1';
				bpi_rd_data <= (others => '0'); 
				last_state <= idle;
		   	
		elsif rising_edge(clk) then
		
			-- Assign the current state to the last state.
			last_state <= bpi_state;
		
			-- Toggle the address enable before every write or read transaction (pw >= 7 ns).
			if bpi_state = add then
				bpi_adv_b <= '0';
			else
				bpi_adv_b <= '1';
			end if;
				
			-- Activate write enable to complete a write transaction (pw >= 40 ns).
			if bpi_state = wr then
				bpi_fwe_b <= '0';
			else
				bpi_fwe_b <= '1';
			end if;
				
			-- Activate output enable to complete a read transaction (pw >= 96 ns).
			if bpi_state = rd then
				bpi_foe_b <= '0';
				-- Register the input (read) data from the flash.
				bpi_rd_data <= bpi_d_in;
			else
				bpi_foe_b <= '1';
			end if;
				
			-- Tristate the FPGA data bus driver during bus turn around and read states.
			if bpi_state = bus_turn or bpi_state = rd then
				bpi_d_tristate <= '1';
			else
				bpi_d_tristate <= '0';
			end if;
				
			-- Activate chip select for any type of transaction.
			if bpi_state = idle then
				bpi_fcs_b <= '1';
			else
				bpi_fcs_b <= '0';
			end if;
				
		end if;
	end process;
	
--************************************************** EXTERNAL PIN ASSIGNMENTS **************************************************

	-- Assign the address bus.
	bpi_a <= address;
	
	-- Assign the revision select bits.
	bpi_rs <= revision_select;
	bpi_rs_disable <= not(revision_enable);
	
	-- Assign the output (write) data bus.
	bpi_d_out <= bpi_wr_data;
	
	-- Generate a busy flag throughout the complete bus transaction.
    bpi_busy <= '0' when bpi_state = idle else '1';
	
	-- This control is a hardware build option that is only present if requested.
	bpi_vpp <= program_voltage;

end structural;
