--*******************************************************************************************************************************
--
-- Description: This core contains a decoder that interfaces to the slave interface of the PLDA EZDMA core.  A valid write
--				transaction will set the corresponding BARx_write flag that is output from the core.  A valid read transaction
--				will program DMA Channel 0 of the PLDA EZDMA core to perform a Completion (with data) transaction.  Please note
--				that DMA Channel 0 is used to complete slave read requests.
--
--				This core also contains the logic to multiplex the correct data source into the dma_rddata port of the PLDA
--				EZDMA core based on the state of the dma_rdchannel input.
--					
-- Rev History: 09-04-2013:	Initial release.
--
--*******************************************************************************************************************************

library ieee;
use ieee.std_logic_1164.all;

use work.model_specific_package.all;

entity pci_transaction_decoder is
	port (
	-- Reset and clocks.
	rst         	: in std_logic;									--	Active high asynchronous reset.
	clk 			: in std_logic;									--	Bus clock.

	-- Connections to the slave interface ports on the PLDA EZDMA2 core.
	slv_bar0 		: in std_logic;									-- Selects the BAR space for the current slave request.
	slv_bar2 		: in std_logic;									-- Selects the BAR space for the current slave request.
	slv_addr 		: in std_logic_vector(15 downto 0);				-- Selects the local address for the current slave request.
	slv_readreq 	: in std_logic;									-- Logic one indicates that the current slave request is a read transaction.
	slv_writereq 	: in std_logic;									-- Logic one indicates that the current slave request is a write transaction.
	slv_write 		: in std_logic;									-- Logic one indicates that data must be written to the current address.
	slv_bytecount 	: in std_logic_vector(12 downto 0);				-- Indicates the size of the current transaction in bytes.
	slv_cpladdr 	: in std_logic_vector(31 downto 0);				-- Conveys the DMA Completion address required to finish a read transaction.
	slv_cplparam 	: in std_logic_vector(7 downto 0);				-- Conveys the DMA Completion parameters required to finish a read transaction.
	slv_accept 		: out std_logic;								-- Logic one indicates that the current request is accepted.
	slv_abort 		: out std_logic;								-- Logic one indicates that the current request is aborted.

	-- Connections to the DMA Channel 0 control ports on the PLDA EZDMA2 core.
	dma_regin		: out std_logic_vector(127 downto 0);			-- Configuraton settings for the PLDA EZDMA2 core.
	dma_param 		: out std_logic_vector(23 downto 0);			-- DMA parameter settings for the PLDA EZDMA2 core.
	dma_control 	: out std_logic_vector(5 downto 0);				-- DMA control settings for the PLDA EZDMA2 core.  
	dma_status 		: in  std_logic;								-- DMA status from the PLDA EZDMA2 core.           
	dma_fifocnt 	: out std_logic_vector(12 downto 0);			-- FIFO count into the PLDA EZDMA2 core (not used).
	
	-- Connections to the PLDA EZDMA2 core DMA controller ports with the same name.
	dma_rdchannel 	: in std_logic_vector(7 downto 0);				-- Channel select bus from the PLDA EZDMA2 core used for DMA write transactions.
	dma_rddata_mux 	: out std_logic_vector(AXI_WIDTH-1 downto 0);	-- Data bus used for DMA write transactions through the PLDA EZDMA2 core.
	dma_rd 			: in std_logic;						
	
	-- Connections to the BAR registers that service target (slave) transactions.
	bar0_write 		: out std_logic;								-- Logic one indicates that the current write request is directed at BAR0.
	bar2_write 		: out std_logic;								-- Logic one indicates that the current write request is directed at BAR2.
	bar0_rddata 	: in std_logic_vector(AXI_WIDTH-1 downto 0);	-- BAR0 local data bus used for DMA Channel 0 Completions to a read request.
	bar2_rddata 	: in std_logic_vector(AXI_WIDTH-1 downto 0);	-- BAR2 local data bus used for DMA Channel 0 Completions to a read request.
	bar0_read		: out std_logic;								-- Logic one indicates that the current read request is directed at BAR0.
	bar2_read		: out std_logic;								-- Logic one indicates that the current read request is directed at BAR2.
	bar0_read_latency 	: in std_logic_vector(1 downto 0);			-- Read latency from BAR0 as defined by the PLDA EZDMA2 core.
	bar2_read_latency 	: in std_logic_vector(1 downto 0);			-- Read latency from BAR2 as defined by the PLDA EZDMA2 core.
	
	-- Data supplied from multiple channels for DMA write transactions.
	-- Note:  Unused busses can be tied to logic zero.
	dma1_rddata 	: in std_logic_vector(AXI_WIDTH-1 downto 0);	-- DMA Channel 1 local data bus used for DMA write transactions.
	dma2_rddata 	: in std_logic_vector(AXI_WIDTH-1 downto 0);	-- DMA Channel 2 local data bus used for DMA write transactions.
	dma3_rddata 	: in std_logic_vector(AXI_WIDTH-1 downto 0);	-- DMA Channel 3 local data bus used for DMA write transactions.
	dma4_rddata 	: in std_logic_vector(AXI_WIDTH-1 downto 0);	-- DMA Channel 4 local data bus used for DMA write transactions.
	dma5_rddata 	: in std_logic_vector(AXI_WIDTH-1 downto 0);	-- DMA Channel 5 local data bus used for DMA write transactions.
	dma6_rddata 	: in std_logic_vector(AXI_WIDTH-1 downto 0);	-- DMA Channel 6 local data bus used for DMA write transactions.
	dma7_rddata 	: in std_logic_vector(AXI_WIDTH-1 downto 0)		-- DMA Channel 7 local data bus used for DMA write transactions.
	);
end pci_transaction_decoder;

architecture structural of pci_transaction_decoder is

	signal cpl_readreq		: std_logic;					-- Flag to indicate that DMA Channel 0 is available to perform a Completion in response to a read request. 
	signal req_outstanding	: std_logic;					-- Flag to indicate that there is a Completion request outstanding on DMA Channel 0. 
	signal bar_store		: std_logic_vector(1 downto 0);	-- Register to store the BAR space requested in the last slave read transaction. 
	
begin

--************************************************* TARGET TRANSACTION DECODER *************************************************
	
	-- Check the status of DMA Channel 0 to determine whether a read request can be processed as a Completion.
	cpl_readreq <='1' when slv_readreq = '1' and dma_status = '0' and req_outstanding = '0' else '0';
	
	-- Accept all write requests to BAR0 and BAR2 and read requests based on the status of DMA Channel 0.
	slv_accept <= '1' when (slv_bar0 = '1' or slv_bar2 = '1') and (slv_writereq = '1' or cpl_readreq = '1') else '0';
	
	-- Abort any write requests outside of BAR0 or BAR2 and any read requests while DMA Channel 0 is busy.
	slv_abort <= '1' when (slv_bar0 = '0' and slv_bar2 = '0') and (slv_writereq = '1' or cpl_readreq = '1') else '0';

	-- Set the flags to indicate that a valid write transaction is in process.
	bar0_write <= slv_write and slv_bar0;
	bar2_write <= slv_write and slv_bar2;
	
--**************************************************** READ COMPLETION DMA *****************************************************

	-- The local address is copied from the slave read request, only 16 bits are used since the largest BAR is 64 kbytes.
    dma_regin(127 downto 96) <= x"0000" & "00" & slv_addr(13 downto 0) when slv_bar0 = '1' else x"0000" & slv_addr;
    
	-- The transfer size is copied from the slave read request.
	dma_regin(95 downto 64) <= x"0000" & "000" & slv_bytecount;
    
	-- The PCI Completion address is copied from the slave read request.
	dma_regin(63 downto 0) <= x"00000000" & slv_cpladdr;

	-- Set the DMA Completion parameters.
	dma_param(23 downto 12) <= slv_cplparam & "0000";	-- Parameters copied from the read request
	dma_param(11 downto 8)  <= "0100";        			-- Completion command
	dma_param(7 downto 4)   <= "0000"; 				-- Byte enables do not apply
	dma_param(3 downto 2)   <= bar0_read_latency when slv_bar0 = '1' else bar2_read_latency; 	-- Latency = 0 (data available on next clock cycle)
	dma_param(1 downto 0)   <= "01"; 					-- Select RAM mode

	-- Program the DMA channel to perform a Completion as soon as the read request is received.
    dma_control <= "011111" when (slv_bar0 = '1' or slv_bar2 = '1') and cpl_readreq = '1' else (others=>'0');
    
	-- Store the BAR that is associated with the last read Completion request since it is not provided by the DMA core.
	process (clk,rst)
	begin
		if rst = '1' then
			bar_store <= (others=>'0');
			req_outstanding <= '0';
		elsif rising_edge (clk) then
			if ((slv_bar0 = '1' or slv_bar2 = '1') and cpl_readreq = '1') then
				bar_store <= slv_bar2 & slv_bar0;
				req_outstanding <= '1';
			elsif ((dma_rdchannel = x"01") and (dma_rd = '1')) then
				req_outstanding <= '0';
			end if;
		end if;
	end process;

	-- Assign the contents of BAR_store to output ports.
	bar0_read <= bar_store(0);
	bar2_read <= bar_store(1);
	
	-- The FIFO count is not used because this DMA channel operates in RAM mode.
	dma_fifocnt <= (others=>'0');
	
--**************************************************** DMA DATA BUS ROUTING ****************************************************

--	process (dma_rdchannel, bar_store, bar0_rddata, bar2_rddata, dma1_rddata, dma2_rddata, dma3_rddata, dma4_rddata, dma5_rddata, dma6_rddata, dma7_rddata)
	process (clk,rst)
	begin
		if rst = '1' then
			dma_rddata_mux <= (others=>'0');
		elsif rising_edge (clk) then
			-- Select the DMA input data bus based on the channel requested by the PLDA EZDMA2 controller.
			case dma_rdchannel is
				when x"01" =>
					-- Decode the selected BAR from the lowest two address bits.
					if bar_store(0) = '1' then
						dma_rddata_mux <= bar0_rddata;
					elsif bar_store(1) = '1' then
						dma_rddata_mux <= bar2_rddata;
					end if;
				when x"02" => dma_rddata_mux <= dma1_rddata;
				when x"04" => dma_rddata_mux <= dma2_rddata;
				when x"08" => dma_rddata_mux <= dma3_rddata;
				when x"10" => dma_rddata_mux <= dma4_rddata;
				when x"20" => dma_rddata_mux <= dma5_rddata;
				when x"40" => dma_rddata_mux <= dma6_rddata;
				when x"80" => dma_rddata_mux <= dma7_rddata;
				when others => null;
			end case;
		end if;
	end process;
	
end structural;
